/*
 * fwatch: watch files.
 * 92/09/02 john h. dubois iii  spcecdt@armory.com
 * 96/05/07 Added fMx options.
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <memory.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>
#include <sys/timeb.h>
#include <time.h>

#define WriteString(S,fd) write(fd,S,strlen(S))
#define ErrExit(Message) WriteString(Message,2),exit(1)

char Version[] = "@(#) fwatch 1.1 96/05/07";

char HelpMsg[] = "Usage: "\
"fwatch [-acme] -f<timeformat> -M<milliseconds> -<seconds> filename ...\n"\
"fwatch watches files for time changes.\n" \
"At intervals, it does a stat(S) on each file, and prints on the standard\n"\
"output the names of those that have had an inode time change (up or\n"\
"down).  One name is printed per line.  By default, only the modification\n"\
"time is checked.  The default stat interval is one second.  fwatch\n"\
"continues executing until killed by a signal.  By default, fwatch will\n"\
"abort with an error message and nonzero exit value if a file cannot be\n"\
"stat()ed.\n"\
"Options:\n"\
"Note: if an option takes a value, the value must be given immediately\n"\
"after the option, in the same argument.\n"\
"-e: Do not print an error message or exit if a file cannot be stat()ed.\n"\
"    The name of a file is printed if it disappears or can no longer be\n"\
"    accessed for any other reason, and it is printed again if it comes back.\n"\
"-f<timeformat>: Before each filename, print the current time.  If\n"\
"    <timeformat> is non-null, it is used as a time format string (see\n"\
"    strftime(S)).  If it is null, the UNIX epoch time, in seconds, is\n"\
"    printed as a fixed-point value with two fractional digits.\n"\
"If any of the following three options is given, each of the indicated\n"\
"times will be checked and the filename will be printed if any of them\n"\
"have changed.  The default is equivalent to given only -m.\n"\
"-m: Check the modify time.\n"\
"-a: Check the access time.\n"\
"-c: Check the creation (inode change) time.\n"\
"One or both of the following options may be given; if both, the time\n"\
"between checks is the sum of both times:\n"\
"-<seconds>: Check file times every <seconds> seconds.\n"\
"-M<milliseconds>: Check file times every <milliseconds> milliseconds.\n"\
"    The value given is silently rounded up to the next multiple of the\n"\
"    system clock interval, typically 10 milliseconds.  Note that if file\n"\
"    times are being updated by normal means, file times cannot change more\n"\
"    than once per second (since that is the resolution of the file times),\n"\
"    so if only normal time updates are being checked for, there is little\n"\
"    point in checking more than once per second.\n";

int ChkAccess, ChkMod, ChkCreate;
int Seconds = 1;
int Milliseconds = 0;
int NoErr = 0;
char *timeFormat = (char *) 0;
int Debug = 0;

main(int argc, char *argv[])
{
    char *Arg;
    int i;

    for (i = 1; i < argc && *(argv[i]) == '-'; i++)
	for (Arg = argv[i] + 1; *Arg; Arg && Arg++)
	    switch (*Arg) {
	    case 'a': 
		ChkAccess = 1;
		break;
	    case 'm': 
	        ChkMod = 1;
		break;
	    case 'c': 
		ChkCreate = 1;
		break;
	    case 'e': 
		NoErr = 1;
		break;
	    case '1': case '2': case '3': case '4': case '5':
	    case '6': case '7': case '8': case '9':
		Seconds = atoi(Arg);
		if (Seconds < 1)
		    ErrExit("Must give a positive value for sleep time.\n");
		Milliseconds = 0;
		Arg = (char *) 0;
		break;
	    case 'M':
		Milliseconds = atoi(Arg+1);
		if (Milliseconds < 1)
		    ErrExit("Must give a positive value with -M.\n");
		Seconds = 0;
		Arg = (char *) 0;
		break;
	    case 'f':
		timeFormat = Arg+1;
		Arg = (char *) 0;
		break;
	    case 'h': 
		WriteString(HelpMsg,1);
		exit(0);
	    case 'x':
		Debug = 1;
		break;
	    default:
		ErrExit("Invalid flag.  Use -h for help.\n");
	    }
	    if (!ChkAccess && !ChkMod && !ChkCreate)
		ChkMod = 1;
    argv += i;
    argc -= i;
    if (!argc)
	ErrExit(HelpMsg);
    CheckFiles(argc,argv);
}

int CheckFiles(int Num, char *Files[])
{
    struct stat *Stats1, *Stats2, *Tmp;

    if (!(Stats1 = calloc(Num,sizeof(struct stat))) ||
    !(Stats2 = calloc(Num,sizeof(struct stat))))
	ErrExit("Not enough memory.\n");
    StatFiles(Num,Files,Stats1);
    while (1) {
	Tmp = Stats1;
	Stats1 = Stats2;
	Stats2 = Tmp;
	StatFiles(Num,Files,Stats1);
	CompareStats(Num,Stats1,Stats2,Files);
	if (Seconds) {
	    if (Debug)
		WriteString("sleeping...\n",1);
	    sleep(Seconds);
	}
	if (Milliseconds) {
	    if (Debug)
		WriteString("napping...\n",1);
	    nap(Milliseconds);
	}
    }
}

CompareStats(int Num, struct stat *Stats1, struct stat *Stats2,char *Files[])
{
    int i;
    char Buf[81];
    struct timeb Time;
    char s[11];
    char ms[3];

    for (i = 0; i < Num; i++) {
	if (ChkAccess && Stats1 -> st_atime != Stats2 -> st_atime ||
	ChkMod && Stats1 -> st_mtime != Stats2 -> st_mtime ||
	ChkCreate && Stats1 -> st_ctime != Stats2 -> st_ctime) {
	    if (timeFormat) {
		ftime(&Time);
		if (*timeFormat) {
		    strftime(Buf,80,timeFormat,localtime(&Time.time));
		    WriteStrings(1,Buf," ",(char *)0);
		}
		else {
		    ltoa(Time.time,s,10);
		    ms[0] = '0' + Time.millitm / 100;
		    ms[1] = '0' + (Time.millitm / 10) % 10;
		    ms[2] = '\0';
		    WriteStrings(1,s,".",ms," ",(char *)0);
		}
	    }
	    WriteStrings(1,Files[i],"\n",(char *)0);
	}
	Stats1++;
	Stats2++;
    }
}

StatFiles(int Num, char **Files, struct stat *Stats)
{
    static struct stat ErrorStat;
    int i;

    for (i = 0; i < Num; i++)
	if (stat(Files[i],Stats + i) == -1)
	    if (NoErr)
		Stats[i] = ErrorStat;
	    else {
		WriteStrings(2,"fwatch: ",Files[i],": ",strerror(errno),
		"\n",(char *) 0);
		exit(1);
	    }
}

WriteStrings(int fd, ...)
{
    va_list ap;
    char *S;

    va_start(ap,fd);
    while (S = va_arg(ap,char *))
	write(fd,S,strlen(S));
    va_end(ap);
}
